(** Types and functions for manipulation on HTTP headers. *)
module Headers : sig
  type t

  val show: t -> string

  (** Creates new empty headers object *)
  val empty: t

  (** Creates new headers object from a given list.

      NOTE: If the list contains duplicate keys, only the last value will be
      kept. Keys are case-insensitive, 'FOO' and 'foo' considered equal.
   *)
  val of_list: (string * string) list -> t

  (** Converts headers to a list *)
  val to_list: t -> (string * string) list

  (** Returns the value of the header [~key] if it exists *)
  val find: t -> key:string -> string option

  (** Sets the value of the header [~key] to [~value]*)
  val set: t -> key:string -> value:string -> t

  (** Removes the header [~key] *)
  val remove: t -> key:string -> t
end

(** Types and functions representing body of HTTP request or response *)
module Body : sig
  type t

  val show: t -> string

  (** Creates new empty body *)
  val empty: t

  (** Creates new body from a given string *)
  val of_string: string -> t

  (** Creates new body from a given Lwt stream *)
  val of_stream: string Lwt_stream.t -> t

  (** Returns the whole body if it was given in a single chunk, e.g.
      constructed using [of_string] or [empty] functions. Returns [None] if the
      body is streamed, e.g. constructed using [of_stream] function or come from
      actual request/response with [Transfer-Encoding: chunked] header.

      Use [read_up_to] to read the streamed body or [to_stream] and [Lwt_stream]
      module functions to process it chunk by chunk.
  *)
  val to_string: t -> string option

  (** Returns Lwt stream containing the body *)
  val to_stream: t -> string Lwt_stream.t

  (** Reads up to [~max_bytes] bytes (default: 1MB) from the body.

      IMPORTANT: the actual size could differ from [~max_bytes] because body is
      read in chunks, not byte by byte.
  *)
  val read_up_to: ?max_bytes:int -> t -> string Lwt.t

  (** Returns the length of the body in bytes if it is known, e.g. the body is
      not streamed *)
  val length: t -> int option
end

(** Types and functions for accessing urlencoded form body *)
module Form : sig
  type t

  (** Reads and decodes given body constructing a form from it *)
  val of_body: ?max_bytes:int -> Body.t -> t Lwt.t

  (** Returns all form params associated with [~key] *)
  val find_params: t -> key:string -> string list option

  (** Returns the first form param associated with [~key] *)
  val find_param: t -> key:string -> string option
end

(** Types and functions for HTTP request manipilation *)
module Request : sig
  type t

  val show: t -> string

  (** Returns remote address of the client making the request or None if request
      is created manually. *)
  val get_client_addr: t -> Unix.sockaddr option

  (** Returns HTTP method in all caps, e.g. GET, POST, etc *)
  val get_http_method: t -> string

  (** Returns URI of the request *)
  val get_uri: t -> Uri.t

  (** Returns path part of the request URL as a list of chunks split by '/' *)
  val get_path: t -> string list

  (** Returns query string of the request URL as a list of key/values pairs *)
  val get_query_string: t -> (string * string list) list

  (** Returns headers of the request *)
  val get_headers: t -> Headers.t

  (** Returns body of the request *)
  val get_body: t -> Body.t

  (** Returns all query string params associated with [~key] *)
  val find_query_params: t -> key:string -> string list option

  (** Returns the first query string param associated with [~key] *)
  val find_query_param: t -> key:string -> string option

  (** Returns the value of the header [~key] if it exists *)
  val find_header: t -> key:string -> string option

  (** Creates new HTTP request with a given [~http_method] and [~uri] *)
  val make: http_method:string -> uri:string -> t

  (** Creates new GET HTTP request for a given URI *)
  val get: string -> t

  (** Creates new POST HTTP request for a given URI *)
  val post: string -> t

  (** Overwrites the body of the request with a given one *)
  val with_body: Body.t -> t -> t

  (** Overwrites the body of the request with a given string *)
  val with_string_body: string -> t -> t

  (** Overwrites the list of headers with a given one *)
  val with_headers: Headers.t -> t -> t

  (** Overwrites the header of the request with a given value. *)
  val with_header: key:string -> value:string -> t -> t
end

(** Type and functions for HTTP response manipulation *)
module Response : sig
  type t

  val show: t -> string

  (** Returns status of the response *)
  val get_status: t -> int

  (** Returns headers of the response *)
  val get_headers: t -> Headers.t

  (** Returns body of the response *)
  val get_body: t -> Body.t

  (** Returns the value of the header [~key] if it exists *)
  val find_header: t -> key:string -> string option

  (** Returns an empty 200 response *)
  val ok: t

  (** Returns an empty response with a given status code *)
  val of_status: int -> t

  (** Returns a 301 response redirecting to [~url].
      Response body is set to an HTML page containing a link to [~url]. *)
  val redirect: url:string -> t

  (** Overwrites the body of the response with a given one. *)
  val with_body: Body.t -> t -> t

  (** Sets the body of the response to a given string. *)
  val with_string_body: string -> t -> t

  (** Overwrites the list of headers with a given one *)
  val with_headers: Headers.t -> t -> t

  (** Sets the header of the response to a given value. *)
  val with_header: key:string -> value:string -> t -> t
end

(** Client-side functions *)
module Client : sig
  (** Issues an HTTP request using information provided in a given request
      object and returns a response from the server. *)
  val send: Request.t -> Response.t Lwt.t
end

(** Type of server-side handler that processes requests *)
module Handler : sig
  type t = Request.t -> Response.t Lwt.t
end

(** Type of middleware that checks and transforms requests and passes them down
    the chain *)
module Middleware : sig
  type t = Handler.t -> Handler.t
end

(** Server-side functions *)
module Server : sig
  (** Starts a new HTTP server listening on given [~address] and [~port].
      Each request is processed with a given handler. Optional [?on_exn]
      parameter is a function called whenever an exception is raised inside the
      handler (default is to log the exception to stderr). *)
  val listen: ?on_exn:(exn -> unit) -> address:string -> port:int -> Handler.t -> unit Lwt.t
end
